from pyspark import SparkContext
from pyspark.mllib.linalg import Vectors
from pyspark.mllib.clustering import KMeans
import numpy as np
import scipy as scp
import scipy.sparse as sps
import scipy.io as sio
from scipy.optimize import newton_krylov
from scipy.optimize import fmin_l_bfgs_b
from scipy.stats import norm as normal
import pyroc
import sys
import os
from kernel import *
from model import *
from gradient_descent import *


# evalute predictive performance on test data
# data.ind , tensor indices for test values
# data.y, the ground-truth tensor entry
# model.U, the learned latent factors
# model.V, the learned pseudo input
# model.post_mu, the learned posterior mean of pseudo input
# model.Kmm, kernel matrix of the psuedo input
def predict_accuracy(model, data, ker_param):
    if model.task == 'regression':
        name = 'mse'
    else:
        name = 'auc'
    V = model.V
    U = model.U
    nmod = len(U)
    ind = data.ind
    y = data.y
    part = [None for k in xrange(nmod)]
    for k in xrange(nmod):
        part[k] = U[k][flat(ind[:, k]), :]
    X = np.hstack(part)
    Knm = kernel_cross_func(X, V, ker_param)
    pred = Knm * np.linalg.solve(model.Kmm, model.post_mu)
    if model.task == 'regression':
        val = np.mean((pred - y).A ** 2)
    else:
        roc = pyroc.ROCData([(y[i], pred[i]) for i in range(len(y)) ])
        val = roc.auc()
    print '%s = %f\n' % (name, val)
    return val


''' read latent factors from a resource file.
    <row_index>\t col1\t col2\t col3 ...
'''


def load_latent_factors(sc, file_name, n, k, flipsign=True):
    data = sc.textFile(file_name).collect()
    U = np.zeros([n, k])
    for line in data:
        terms = line.split('\t')
        row, val = int(terms[0]), np.array([float(terms[i]) for i in range(1, len(terms))])
        U[row, :] = val
    # flip the sign
    if flipsign:
        loc = np.argmax(np.abs(U), 0)
        for r in xrange(k):
            if U[loc[r], r] < 0:
                U[:, r] *= -1

    return np.mat(U)


# mapping process to generate input for kmeans to generate pseudo input init
def map4kmeans(line, U):
    terms = line.split('\t')
    ind = np.array([int(terms[i]) for i in range(0, len(terms) - 1)]).astype(int)
    nmod = len(U)
    part = [None for k in xrange(nmod)]
    for k in xrange(nmod):
        part[k] = U[k].A[ind[k], :]
    x = np.hstack(part)
    return x


# convert a  matrix to a flat array
def flat(b): return b.A.reshape(b.size)



def map_binary_init_lam(rdd, U, V, kp):
    nmod = len(U)
    ker_param = KernelParam()
    ker_param.recover('ard', kp)
    ind,y,u2ind = rdd
    #assemble X
    n = ind.shape[0]
    part = [None for k in xrange(nmod)]
    for k in xrange(nmod):
        part[k] = U[k][flat(ind[:,k]),:]
    X = np.hstack(part)
    Kmn = kernel_cross_func(V, X, ker_param)
    C = Kmn*(1.2876*(2.0*y-1.0))
    return C


# generate ind, y, u2ind for each partition
def do_map_partition(rdd_iter, nvec):
    ind_list = []
    y = []
    for line in rdd_iter:
        terms = line.split('\t')
        ind = np.mat([int(terms[i]) for i in range(0, len(terms) - 1)])
        val = float(terms[-1])
        ind_list.append(ind)
        y.append(val)

    '''
    if len(ind_list)==0:
        return
    '''
    indmat = np.vstack(ind_list)
    n, nmod = indmat.shape
    elements = np.ones(n)
    u2ind = [None for j in xrange(nmod)]
    col = np.array(range(n))
    for j in xrange(nmod):
        row = indmat[:, j].A.reshape(n)
        u2ind[j] = sps.coo_matrix((elements, (row, col)), shape=(nvec[j], n))

    yield [indmat, np.mat(y).T, u2ind]


# Only for continuous data: generate C, D
def map_process_continuous(rdd, U, V, kp):
    nmod = len(U)
    ker_param = KernelParam(type='ard')
    ker_param.recover('ard', kp)
    ind, y, u2ind = rdd
    # assemble X
    n = ind.shape[0]
    part = [None for k in xrange(nmod)]
    for k in xrange(nmod):
        part[k] = U[k][flat(ind[:, k]), :]
    X = np.hstack(part)
    Kmn = kernel_cross_func(V, X, ker_param)
    C = Kmn * y
    D = Kmn * Kmn.T
    # print y.shape[0], 'debug'
    return [C, D]


# Only for binary data
# generate D
def map_process_binary_step1(rdd, U, V, kp):
    nmod = len(U)
    ker_param = KernelParam(type='ard')
    ker_param.recover('ard', kp)
    ind, y, u2ind = rdd
    # assemble X
    n = ind.shape[0]
    part = [None for k in xrange(nmod)]
    for k in xrange(nmod):
        part[k] = U[k][flat(ind[:, k]), :]
    X = np.hstack(part)
    Kmn = kernel_cross_func(V, X, ker_param)
    D = Kmn * Kmn.T
    return D



#x, the mean of a before truncation; y, the observed label; tau, inverse of variance
#x,y are flatten arrays, tau can be sclar or flatten array
def probit_mean(x, y, tau = 1):
    xt = x*np.sqrt(tau)
    return x + 1.0/np.sqrt(tau)*(2*y-1)*normal.pdf(xt)/normal.cdf((2*y-1)*xt)






# the mean of the a^2 where a is probit latent variable
# x, the mean of a; y, the observed label
# x,y are flatten arrays
def probit_square_mean(x, y):
    return x ** 2 + 1 + (2 * y - 1) * x * normal.pdf(x) / normal.cdf((2 * y - 1) * x)


# the entropy of a probit dist.,
# b, the mean; y, the observed label
# b, y are flatten arrays
def probit_entropy(b, y):
    n = y.size
    z = normal.cdf((2 * y - 1) * b)
    return 0.5 * n * (np.log(2 * np.pi) + 1) + np.sum(np.log(z)) - np.sum((2 * y - 1) * b * normal.pdf(b) / (2 * z))

#for fixed point iteration of \lambda
def map_inner_lam(rdd, U, V, kp, lam):
    nmod = len(U)
    ker_param = KernelParam(type='ard')
    ker_param.recover('ard', kp)
    ind, y, u2ind = rdd
    # assemble X
    n = ind.shape[0]
    part = [None for k in xrange(nmod)]
    for k in xrange(nmod):
        part[k] = U[k][flat(ind[:, k]), :]
    X = np.hstack(part)
    Kmn = kernel_cross_func(V, X, ker_param)
    mu = Kmn.T*lam
    b = (2*flat(y)-1)*normal.pdf(flat(mu))/normal.cdf((2*flat(y)-1)*flat(mu))
    b = np.mat(b.reshape([n,1]))
    C = Kmn*b
    return C



# generate C, b, b^2
def map_process_binary_step2(rdd, U, V, kp, lam):
    nmod = len(U)
    ker_param = KernelParam(type='ard')
    ker_param.recover('ard', kp)
    ind, y, u2ind = rdd
    # assemble X
    n = ind.shape[0]
    part = [None for k in xrange(nmod)]
    for k in xrange(nmod):
        part[k] = U[k][flat(ind[:, k]), :]
    X = np.hstack(part)
    Kmn = kernel_cross_func(V, X, ker_param)
    # compute q(b), <b>, <b^2>, H(q(b))
    mu = Kmn.T*lam
    b = probit_mean(flat(mu), flat(y))
    b = np.mat(b.reshape([n,1]))
    b2 = np.sum(probit_square_mean(flat(mu), flat(y)))
    H = probit_entropy(flat(mu), flat(y));
    #computer other statistics
    C = Kmn*b
    return [C, b2, H, ind, b, u2ind]


# for continuous: input [indmat, y, u2ind]
# for binary: input [indmat, b, u2ind]
def map_part_grad(rdd, U, V, kp, F, G):
    ind, y, u2ind = rdd
    ker_param = KernelParam(type='ard')
    ker_param.recover('ard', kp)
    nmod = len(U)
    part = [None for k in xrange(nmod)]
    for k in xrange(nmod):
        part[k] = U[k][flat(ind[:, k]), :]
    X = np.hstack(part)
    Kmn = kernel_cross_func(V, X, ker_param)
    G2 = y * F.T + Kmn.T * G

    [gp, gkp] = ker_cross_grad(V, X, G2, Kmn, ker_param)
    gX = ker_cross_grad(X, V, G2.T, Kmn.T, ker_param, calculate_gradient_kernel_params=False)
    gU = [None for k in xrange(nmod)]
    st = 0
    for k in xrange(nmod):
        dim = U[k].shape[1]
        gU[k] = u2ind[k].dot(gX[:, st:st + dim])
        st = st + dim
    # print 'debug', ind.shape[0]
    return [gp, gU, gkp]


def logdet(A):
    (s, lg) = np.linalg.slogdet(A)
    return lg


# x: input vector, an np.array
# sc, sparkcontext object
def negative_log_probability_and_gradient(sc, tensor_rdds, x, model, ker_param):
    V = np.mat(x[0:model.pseudo_num * model.pseudo_dim].reshape([model.pseudo_num, model.pseudo_dim]))
    x = x[model.pseudo_num * model.pseudo_dim:]
    U = [None for i in range(model.mod_num)]
    st = 0
    for k in range(model.mod_num):
        U[k] = np.mat(x[st:st + model.tensor_size[k] * model.dim[k]].reshape([model.tensor_size[k], model.dim[k]]))
        st = st + model.tensor_size[k] * model.dim[k]
    xu = x[0:st]
    xkp = x[st:]
    # to ensure the postiveness of kernel params, the func. defined over f(|x|)
    if model.task == 'regression':
        model.noise = np.abs(xkp[-1])
        ker_param.w = np.abs(xkp[-2])
        ker_param.gamma = np.abs(xkp[0:-2])
    else:
        # for classification, we do not need to estimate beta---it is fixed to 1
        ker_param.w = np.abs(xkp[-1])
        ker_param.gamma = np.abs(xkp[0:-1])

    bU = sc.broadcast(U)
    bker_param_vec = sc.broadcast(ker_param.tovec())
    bV = sc.broadcast(V)

    if model.task == 'regression':
        A = model.y2sum
        B = model.entry_num * (ker_param.w + ker_param.kreg)
        C, D = tensor_rdds.map(
            lambda rdd: map_process_continuous(rdd, bU.value, bV.value, bker_param_vec.value)).reduce(
            lambda x, y: [x[0] + y[0], x[1] + y[1]])
        Kmm = kernel_func(V, ker_param)
        KmmInvD = np.linalg.solve(Kmm, D)
        E = Kmm + model.noise * D
    else:
        D = tensor_rdds.map(lambda rdd: map_process_binary_step1(rdd, bU.value, bV.value, bker_param_vec.value)).reduce(
            lambda x, y: x + y)
        Kmm = kernel_func(V, ker_param)
        E = Kmm + model.noise * D
        C = tensor_rdds.map(lambda rdd:map_binary_init_lam(rdd, bU.value, bV.value, bker_param_vec.value)).reduce(lambda x,y: x+y)
        lam = np.linalg.solve(E, C)
        for iter in xrange(model.inner_iter):
            old_lam = lam.copy()
            bLam = sc.broadcast(lam)
            C = tensor_rdds.map(lambda rdd:map_inner_lam(rdd, bU.value, bV.value, bker_param_vec.value, bLam.value)).reduce(lambda x,y:x+y)
            bLam.unpersist()
            C = D*lam + C
            lam = np.linalg.solve(E, C)
            diff = np.linalg.norm(lam - old_lam)
            if diff < model.inner_tol:
                break
            print 'iter = %d, diff = %g, rel-diff = %g'%(iter, diff, diff/np.linalg.norm(lam))
            

        bLam = sc.broadcast(lam)
        tensor_rdds_enriched = tensor_rdds.map(
            lambda rdd: map_process_binary_step2(rdd, bU.value, bV.value, bker_param_vec.value, bLam.value)).persist()
        # print 'debug', len(tensor_rdds_enriched.first())
        C, A, entropy = tensor_rdds_enriched.map(lambda rdd: rdd[0:3]).reduce(
            lambda x, y: [x[0] + y[0], x[1] + y[1], x[2] + y[2]])
        B = model.entry_num * (ker_param.w + ker_param.kreg)
        KmmInvD = np.linalg.solve(Kmm, D)
        bLam.unpersist()

    F = np.linalg.solve(E, C)
    # generate obj value & gradients
    obj = -0.5 * model.entry_num * np.log(2 * np.pi) + 0.5 * model.entry_num * np.log(model.noise) \
          + 0.5 * logdet(Kmm) - 0.5 * logdet(E) - 0.5 * model.noise * A - 0.5 * model.noise * B \
          + 0.5 * model.noise * np.trace(KmmInvD) \
          + 0.5 * model.noise ** 2 * (C.T * F).flatten()[0, 0] \
          - 0.5 * np.sum(xu ** 2)
    if model.task == 'classification':
        obj += entropy

    KmmInv = np.linalg.inv(Kmm)
    EInv = np.linalg.inv(E)
    FF = F * F.T
    gp, gkp = ker_grad(V, 0.5 * (KmmInv - EInv - model.noise * (KmmInvD * KmmInv) - model.noise ** 2 * FF), Kmm,
                       ker_param)
    G1 = model.noise ** 2 * F
    G2 = -0.5 * model.noise * (EInv - KmmInv + model.noise ** 2 * FF)
    GG2 = G2 + G2.T

    bG1 = sc.broadcast(G1)
    bGG = sc.broadcast(GG2)
    if model.task == 'regression':
        gp2, gU, gkp2 = tensor_rdds.map(
            lambda rdd: map_part_grad(rdd, bU.value, bV.value, bker_param_vec.value, bG1.value, bGG.value)).reduce(
            lambda x, y: [x[0] + y[0], [x[1][k] + y[1][k] for k in xrange(model.mod_num)], x[2] + y[2]])
    else:
        gp2, gU, gkp2 = tensor_rdds_enriched.map(lambda x: x[3:]).map(
            lambda rdd: map_part_grad(rdd, bU.value, bV.value, bker_param_vec.value, bG1.value, bGG.value)).reduce(
            lambda x, y: [x[0] + y[0], [x[1][k] + y[1][k] for k in xrange(model.mod_num)], x[2] + y[2]])
    # release memory
    bU.unpersist()
    bV.unpersist()
    bker_param_vec.unpersist()
    bG1.unpersist()
    bGG.unpersist()
    if model.task == 'classification':
        tensor_rdds_enriched.unpersist()

    gU = [gU[k] - U[k] for k in xrange(model.mod_num)]
    gp = gp + gp2
    gkp = gkp + gkp2
    gkp[-1] -= (0.5 * model.noise * model.entry_num)
    if model.task == 'regression':
        gbeta = 0.5 * model.entry_num / model.noise - 0.5 * np.trace(
            np.linalg.solve(E, D)) - 0.5 * A - 0.5 * B + 0.5 * np.trace(KmmInvD) + model.noise * (C.T * F).flatten()[
            0, 0]
        gkp = np.mat(np.hstack([flat(gkp), gbeta]))
    gkp = flat(gkp) * (2 * (xkp > 0) - 1)
    g = np.hstack([flat(gp)] + [flat(gU[k]) for k in range(model.mod_num)] + [gkp])

    return [-obj, -g]


# tensor_input, the input training data
# U, the init of the latent factors U
# partition_num, number of partitions for the input
# task: 'regression' or 'classification'
def dfnt_train(sc, tensor_input, U, partition_num, task, pnum=100):
    nvec = [x.shape[0] for x in U]
    nmod = len(nvec)
    rank = [x.shape[1] for x in U]
    ker_param = KernelParam(type='ard', gamma=np.ones(np.sum(rank)))
    bU = sc.broadcast(U)
    kmeans_input = sc.textFile(tensor_input, minPartitions=partition_num).map(lambda x: map4kmeans(x, bU.value))
    # clusters = KMeans.train(kmeans_input, 100, runs = 50, initializationMode='random')
    clusters = KMeans.train(kmeans_input, pnum, runs=1, maxIterations=20, initializationMode='k-means||')
    V = np.mat(np.vstack(clusters.centers))
    bU.unpersist()

    tensor_rdds = sc.textFile(tensor_input, minPartitions=partition_num).mapPartitions(
        lambda x: do_map_partition(x, nvec)).persist()
    model = DFNT(nmod=len(U), nvec=nvec, dim=rank, pdim=np.sum(rank), pnum=pnum, beta=1)
    model.task = task
    if task == 'regression':
        model.y2sum = tensor_rdds.map(lambda x: np.sum(np.multiply(x[1], x[1]))).reduce(lambda x, y: x + y)
    else:
        # fixed point iteration for lambda
        model.inner_iter = 20
        model.inner_tol = 1e-3
    model.entry_num = tensor_rdds.map(lambda x: x[1].shape[0]).reduce(lambda x, y: x + y)

    if task == 'regression':
        x = np.hstack(
            [flat(V)] + [flat(U[k]) for k in range(model.mod_num)] + [ker_param.gamma] + [ker_param.w, model.noise])
    else:
        x = np.hstack([flat(V)] + [flat(U[k]) for k in range(model.mod_num)] + [ker_param.gamma] + [ker_param.w])
    [x, f, hist] = simple_gd(lambda x: negative_log_probability_and_gradient(sc, tensor_rdds, x, model, ker_param), x,
                             init_step=1e-2, max_iter=100)
    '''
    x,f,d = fmin_l_bfgs_b(lambda x:negative_log_probability_and_gradient(sc, tensor_rdds, x, model, ker_param), x0=x, args=(), maxiter=100, iprint=2)
    '''

    V = np.mat(x[0:model.pseudo_num * model.pseudo_dim].reshape([model.pseudo_num, model.pseudo_dim]))
    x = x[model.pseudo_num * model.pseudo_dim:]
    U = [None for i in range(model.mod_num)]
    st = 0
    for k in range(model.mod_num):
        U[k] = np.mat(x[st:st + model.tensor_size[k] * model.dim[k]].reshape([model.tensor_size[k], model.dim[k]]))
        st = st + model.tensor_size[k] * model.dim[k]
    xu = x[0:st]
    xkp = x[st:]
    if task == 'regression':
        model.noise = np.abs(xkp[-1])
        ker_param.w = np.abs(xkp[-2])
        ker_param.gamma = np.abs(xkp[0:-2])
    else:
        ker_param.w = np.abs(xkp[-1])
        ker_param.gamma = np.abs(xkp[0:-1])
    model.V = V
    model.U = U
    model.ker_param = ker_param
    Kmm = kernel_func(V, ker_param)
    bU = sc.broadcast(U)
    bker_param_vec = sc.broadcast(ker_param.tovec())
    bV = sc.broadcast(V)
    if task == 'regression':
        C, D = tensor_rdds.map(
            lambda rdd: map_process_continuous(rdd, bU.value, bV.value, bker_param_vec.value)).reduce(
            lambda x, y: [x[0] + y[0], x[1] + y[1]])
        E = Kmm + model.noise * D
    else:
        D = tensor_rdds.map(lambda rdd: map_process_binary_step1(rdd, bU.value, bV.value, bker_param_vec.value)).reduce(
            lambda x, y: x + y)
        E = Kmm + model.noise * D
        C = tensor_rdds.map(lambda rdd:map_binary_init_lam(rdd, bU.value, bV.value, bker_param_vec.value)).reduce(lambda x,y: x+y)
        lam = np.linalg.solve(E, C)
        for iter in xrange(model.inner_iter):
            old_lam = lam.copy()
            bLam = sc.broadcast(lam)
            C = tensor_rdds.map(lambda rdd:map_inner_lam(rdd, bU.value, bV.value, bker_param_vec.value, bLam.value)).reduce(lambda x,y:x+y)
            bLam.unpersist()
            C = D*lam + C
            lam = np.linalg.solve(E, C)
            diff = np.linalg.norm(lam - old_lam)
            if diff < model.inner_tol:
                break
            print 'iter = %d, diff = %g, rel-diff = %g'%(iter, diff, diff/np.linalg.norm(lam))
         

    bU.unpersist()
    bV.unpersist()
    bker_param_vec.unpersist()
    tensor_rdds.unpersist()

    F = np.linalg.solve(E, C)
    model.Kmm = Kmm
    model.post_mu = model.noise * Kmm * F
    model.post_Sigma = Kmm * np.linalg.solve(E, Kmm)

    return model


# test on dblp dataset, size 10000*200*10000
def test_dblp(sc, rank, partition_num):
    nmod = 3
    # load initialization, to be improved by adding args ...
    U = [None for i in range(nmod)]
    U[0] = load_latent_factors(sc, 'dblp-large-u1', 10000, rank)
    U[1] = load_latent_factors(sc, 'dblp-large-u2', 200, rank)
    U[2] = load_latent_factors(sc, 'dblp-large-u3', 10000, rank)
    nvec = [x.shape[0] for x in U]
    model = dfnt_train(sc, 'dblp-large-tensor.txt', U, partition_num, 'classification')
    #model = dfnt_train(sc, 'dblp-large-tensor.txt', U, partition_num, 'regression')
    #model.task = 'classification'
    ker_param = model.ker_param
    auc = np.zeros(50)
    for j in xrange(50):
        data = Tensor()
        data.ind = np.mat(sio.loadmat('dblp.mat')['data']['test'][0][0][0][j][0][0][0]) - 1
        data.y = np.mat(sio.loadmat('dblp.mat')['data']['test'][0][0][0][j][0][0][1])
        auc[j] = predict_accuracy(model, data, ker_param)

    print 'mean auc = %g (%g)\n' % (np.mean(auc), np.std(auc) / np.sqrt(50))
    print ker_param.tovec()


if __name__ == '__main__':
    sc = SparkContext()
    test_dblp(sc, 3, 5)
